from __future__ import annotations
from awesomeversion import AwesomeVersion

__all__ = ["polling", "wsrpc", "central_unit"]

from .polling import CentralUnitPolling
from .wsrpc import (
    CentralUnitWsRPC,
    DeviceConnectionError,
    InvalidMessage,
    ConnectionClosed,
)

from .central_unit import (
    SwitchBeeError,
    SwitchBeeTokenError,
    SwitchBeeDeviceOfflineError,
    CUVersion,
)


def is_wsrpc_api(api: CentralUnitPolling | CentralUnitWsRPC) -> bool:
    assert isinstance(api.version, CUVersion)
    if AwesomeVersion(str(api.version)) >= AwesomeVersion("1.4.6.1"):
        return True
    return False



from __future__ import annotations

import re
from datetime import timedelta
from logging import getLogger
from typing import Any, List
from abc import ABC, abstractmethod

from aiohttp import ClientSession
from switchbee.const import ApiAttribute, ApiCommand, ApiStatus
from switchbee.device import (
    DeviceType,
    HardwareType,
    SwitchBeeBaseDevice,
    SwitchBeeDimmer,
    SwitchBeeGroupSwitch,
    SwitchBeeRollingScenario,
    SwitchBeeScenario,
    SwitchBeeShutter,
    SwitchBeeSomfy,
    SwitchBeeSwitch,
    SwitchBeeThermostat,
    SwitchBeeTimedSwitch,
    SwitchBeeTimerSwitch,
    SwitchBeeTwoWay,
)

from ..utils import timestamp_now

logger = getLogger(__name__)


class SwitchBeeError(Exception):
    pass


class SwitchBeeTokenError(Exception):
    pass


class SwitchBeeDeviceOfflineError(Exception):
    pass


TOKEN_EXPIRATION = int(timedelta(minutes=55).total_seconds()) * 1000

STATE_MAP = [
    DeviceType.Switch,
    DeviceType.GroupSwitch,
    DeviceType.Dimmer,
    DeviceType.Shutter,
    DeviceType.TimedPowerSwitch,
    DeviceType.Thermostat,
    DeviceType.TimedSwitch,
]


class CUVersion:
    def __init__(self, version: str) -> None:
        self.major: int | str = 0
        self.minor: int = 0
        self.revision: int = 0
        self.build: int = 0

        self._initialize(version)

    def _initialize(self, version) -> None:
        if match := re.match(r"(\d+|\S)\.(\d+)\.(\d+)\((\d+)\)$", version):
            self.major = match.group(1)
            self.minor = int(match.group(2))
            self.revision = int(match.group(3))
            self.build = int(match.group(4))

    def __repr__(self) -> str:
        return f"{self.major}.{self.minor}.{self.revision}.{self.build}"


class CentralUnitAPI(ABC):
    _login_count: int = -1  # we don't count the first login
    _token: str | None = None
    _token_expiration: int = 0

    def __init__(
        self, ip_address: str, user: str, password: str, aiohttp_session: ClientSession
    ) -> None:
        self._ip_address: str = ip_address
        self._username: str = user
        self._password: str = password
        self._aiohttp_session: ClientSession = aiohttp_session

        self._mac: str | None = None
        self._unique_id: str | None = None
        self._version: CUVersion | None = None
        self._name: str | None = None
        self._last_conf_change: int = 0
        self._devices_map: dict[
            int,
            SwitchBeeSwitch
            | SwitchBeeGroupSwitch
            | SwitchBeeTimedSwitch
            | SwitchBeeShutter
            | SwitchBeeSomfy
            | SwitchBeeDimmer
            | SwitchBeeThermostat
            | SwitchBeeScenario
            | SwitchBeeRollingScenario
            | SwitchBeeTimerSwitch
            | SwitchBeeTwoWay,
        ] = {}

        self._modules_map: dict[int, set] = {}

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def version(self) -> CUVersion | None:
        return self._version

    @property
    def mac(self) -> str | None:
        return self._mac

    @property
    def unique_id(self) -> str | None:
        return self._unique_id

    @property
    def devices(
        self,
    ) -> dict[
        int,
        SwitchBeeSwitch
        | SwitchBeeGroupSwitch
        | SwitchBeeTimedSwitch
        | SwitchBeeShutter
        | SwitchBeeSomfy
        | SwitchBeeDimmer
        | SwitchBeeThermostat
        | SwitchBeeScenario
        | SwitchBeeRollingScenario
        | SwitchBeeTimerSwitch
        | SwitchBeeTwoWay,
    ]:
        return self._devices_map

    @property
    def last_conf_change(self) -> int:
        return self._last_conf_change

    @property
    def devices_list(
        self,
    ) -> List[SwitchBeeBaseDevice]:
        return list(self._devices_map.values())

    @property
    def reconnect_count(self) -> int:
        return self._login_count

    def module_display(self, unit_id: int) -> str:
        return " and ".join(list(self._modules_map[unit_id]))

    async def login_if_needed(self) -> None:
        if not self._token or (timestamp_now() >= self._token_expiration):
            logger.info(
                "Logging into the Central Unit due to %s",
                "invalid token" if not self._token else " expiry",
            )
            await self._login()

    @abstractmethod
    async def connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def call(
        self,
        command: str | None = None,
        params: dict[str, Any] | int | list | None = None,
        timeout: int = 10,
    ) -> dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    async def _send_request(self, command: str, params: Any = {}) -> dict:
        raise NotImplementedError

    @abstractmethod
    async def _login(self) -> None:
        raise NotImplementedError

    def _update_login(self, resp: dict[str, Any]) -> None:
        self._login_count += 1
        self._token = resp[ApiAttribute.DATA][ApiAttribute.TOKEN]
        # instead of dealing with time synchronization issue, we
        # calculate one hour from now and set it to be the expiration time of the token
        # self._token_expiration = resp[ApiAttribute.DATA][ApiAttribute.EXPIRATION]
        self._token_expiration = timestamp_now() + TOKEN_EXPIRATION

    async def get_configuration(self) -> dict:
        await self.login_if_needed()
        return await self._send_request(ApiCommand.GET_CONF)

    async def get_multiple_states(self, ids: list) -> dict:
        """returns JSON {'status': 'OK', 'data': [{'id': 212, 'state': 'OFF'}, {'id': 343, 'state': 'OFF'}]}"""
        await self.login_if_needed()
        if not len(ids):
            return {}
        return await self._send_request(ApiCommand.GET_MULTI_STATES, ids)

    async def get_state(self, id: int) -> dict:
        """returns JSON {'status': 'OK', 'data': 'OFF'}"""
        await self.login_if_needed()
        return await self._send_request(ApiCommand.GET_STATE, id)

    async def set_state(self, id: int, state: str | int | dict[str, int | str]) -> dict:
        """returns JSON {'status': 'OK', 'data': 'OFF/ON'}"""
        await self.login_if_needed()
        return await self._send_request(
            ApiCommand.OPERATE, {"directive": "SET", "itemId": id, "value": state}
        )

    async def get_stats(self) -> dict:
        """returns {'status': 'OK', 'data': {}} on my unit"""
        await self.login_if_needed()
        return await self._send_request(ApiCommand.STATS)

    async def fetch_configuration(
        self,
        include: list[DeviceType] | None = [],
    ) -> None:
        await self.login_if_needed()
        data = await self.get_configuration()
        if data[ApiAttribute.STATUS] != ApiStatus.OK:
            raise SwitchBeeError

        # clear the old fetched devices
        self._devices_map.clear()
        self._modules_map.clear()
        self._name = data[ApiAttribute.DATA][ApiAttribute.NAME]
        self._version = CUVersion(data[ApiAttribute.DATA][ApiAttribute.VERSION])
        self._mac = data[ApiAttribute.DATA][ApiAttribute.MAC]
        self._unique_id = (
            None
            if ApiAttribute.CU_CODE not in data[ApiAttribute.DATA]
            else data[ApiAttribute.DATA][ApiAttribute.CU_CODE]
        )
        if include is None:
            return

        for zone in data[ApiAttribute.DATA][ApiAttribute.ZONES]:
            for item in zone[ApiAttribute.ITEMS]:
                try:
                    device_type = DeviceType(item[ApiAttribute.TYPE])
                except ValueError:
                    logger.warning(
                        "Unknown device type %s (%s), Skipping",
                        item[ApiAttribute.TYPE],
                        item[ApiAttribute.NAME],
                    )
                    continue
                except KeyError:
                    logger.error(
                        "device %s missing type attribute, Skipping",
                        item[ApiAttribute.NAME],
                    )
                    continue

                if include and device_type not in include:
                    logger.info(
                        "Skipping %s (%s)", device_type.value, item[ApiAttribute.NAME]
                    )
                    continue

                try:
                    device_hw = HardwareType(item[ApiAttribute.HARDWARE])
                except ValueError:
                    logger.warning(
                        "Unknown hardware type %s (%s), Skipping",
                        item[ApiAttribute.HARDWARE],
                        item[ApiAttribute.NAME],
                    )
                    continue
                except KeyError:
                    logger.error(
                        "device %s missing hardware attribute, Skipping",
                        item[ApiAttribute.NAME],
                    )
                    continue

                try:
                    device_id = item[ApiAttribute.ID]
                except KeyError:
                    logger.error(
                        "device %s missing id attribute, Skipping",
                        item,
                    )
                    continue

                try:
                    device_name = item[ApiAttribute.NAME]
                except KeyError:
                    logger.error(
                        "device %s missing name attribute, Skipping",
                        item,
                    )
                    device_name = "Unknown"

                # add switch type device
                if device_type == DeviceType.Switch:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeSwitch(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add dimmer (light) device
                elif device_type == DeviceType.Dimmer:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeDimmer(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add shutter device
                elif device_type == DeviceType.Shutter:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeShutter(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add timed power switch device
                elif device_type == DeviceType.TimedPowerSwitch:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeTimerSwitch(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add scenario
                elif device_type == DeviceType.Scenario:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeScenario(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add group switch only of hardware type != VIRTUAL as we can't read their statuses
                elif (
                    device_type == DeviceType.GroupSwitch
                    and device_hw != HardwareType.Virtual
                ):
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeGroupSwitch(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )

                elif device_type in [DeviceType.Thermostat, DeviceType.VRFAC]:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeThermostat(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                        modes=item[ApiAttribute.MODES],
                        temperature_unit=item[ApiAttribute.TEMPERATURE_UNITS],
                    )

                # add rolling scenario
                elif device_type == DeviceType.RollingScenario:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeRollingScenario(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )

                # add timed switch
                elif device_type == DeviceType.TimedSwitch:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeTimedSwitch(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add two way
                elif device_type == DeviceType.TwoWay:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeTwoWay(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )

                # add somfy
                elif device_type == DeviceType.Somfy:
                    self._devices_map[item[ApiAttribute.ID]] = SwitchBeeSomfy(
                        id=device_id,
                        name=device_name,
                        zone=zone[ApiAttribute.NAME],
                        hardware=device_hw,
                        type=device_type,
                    )
                # add gro

                else:
                    logger.warning(
                        f"Unsupported Type {item[ApiAttribute.TYPE]} {item[ApiAttribute.HARDWARE]}"
                    )
                    continue

                unit_id = self._devices_map[item[ApiAttribute.ID]].unit_id
                if unit_id not in self._modules_map:
                    self._modules_map[unit_id] = set()

                self._modules_map[unit_id].add(
                    self._devices_map[item[ApiAttribute.ID]].hardware.display
                )

    async def fetch_states(
        self,
    ) -> None:
        states = await self.get_multiple_states(
            [
                dev
                for dev in self._devices_map.keys()
                if self._devices_map[dev].hardware != HardwareType.Virtual
                and self._devices_map[dev].type
                in [
                    DeviceType.Switch,
                    DeviceType.GroupSwitch,
                    DeviceType.Dimmer,
                    DeviceType.Shutter,
                    DeviceType.TimedPowerSwitch,
                    DeviceType.Thermostat,
                    DeviceType.TimedSwitch,
                    DeviceType.VRFAC,
                ]
            ]
        )

        if not states:
            logger.debug('get_multiple_states returned empty dict')
            return


        for device_state in states[ApiAttribute.DATA]:
            device_id = device_state[ApiAttribute.ID]
            logger.error(device_state)
            self.update_device_state(device_id, device_state[ApiAttribute.STATE])

    def update_device_state(self, device_id: int, new_state: str | int | dict) -> bool:
        """Update device state."""

        if device_id not in self._devices_map:
            logger.debug("Device id %d is not tracked", device_id)
            return False

        device = self._devices_map[device_id]

        if isinstance(device, SwitchBeeDimmer):
            assert isinstance(new_state, (str, int))
            device.brightness = new_state  # type: ignore

        elif isinstance(device, SwitchBeeShutter):
            assert isinstance(new_state, (str, int))
            device.position = new_state  # type: ignore

        elif isinstance(
            device,
            (
                SwitchBeeSwitch,
                SwitchBeeGroupSwitch,
                SwitchBeeTimedSwitch,
                SwitchBeeTimerSwitch,
            ),
        ):
            assert isinstance(new_state, (int, str))
            device.state = new_state

        elif isinstance(device, SwitchBeeThermostat):
            logger.error(f'Thermostat: {new_state}')
            try:
                assert isinstance(new_state, dict)
                device.state = new_state[ApiAttribute.POWER]
            except TypeError:
                logger.error(
                    "%s: Received invalid state from CU, keeping the old one: %s",
                    device.name,
                    new_state,
                )
                return False

            device.mode = new_state[ApiAttribute.MODE]
            device.fan = new_state[ApiAttribute.FAN]

            device.target_temperature = new_state[ApiAttribute.CONFIGURED_TEMPERATURE]
            device.temperature = new_state[ApiAttribute.ROOM_TEMPERATURE]

        else:
            return False

        return True



from __future__ import annotations

from asyncio import TimeoutError
from datetime import timedelta
from json import JSONDecodeError
from logging import getLogger
from typing import Any
from .central_unit import (
    CentralUnitAPI,
    SwitchBeeError,
    SwitchBeeTokenError,
    SwitchBeeDeviceOfflineError,
)
from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientConnectorError
from switchbee.const import ApiAttribute, ApiCommand, ApiStatus


logger = getLogger(__name__)


class CentralUnitPolling(CentralUnitAPI):
    def __init__(
        self,
        ip_address: str,
        username: str,
        password: str,
        aiohttp_session: ClientSession,
    ) -> None:
        super().__init__(ip_address, username, password, aiohttp_session)

    async def connect(self) -> None:
        await self.fetch_configuration(None)
        await self.fetch_states()

    async def call(
        self,
        command: str | None = None,
        params: dict[str, Any] | int | list | None = None,
        timeout: int = 10,
    ) -> dict[str, Any]:
        """Websocket RPC call."""

        body = {
            ApiAttribute.COMMAND: command,
            ApiAttribute.PARAMS: params,
            ApiAttribute.TOKEN: self._token,
        }

        try:
            async with self._aiohttp_session.post(
                url=f"https://{self._ip_address}/commands", json=body
            ) as response:
                if response.status == 200:
                    try:
                        json_result: dict = await response.json(
                            content_type=None, encoding="utf8"
                        )
                        if json_result[ApiAttribute.STATUS] != ApiStatus.OK:
                            # check if invalid token or token expired
                            if json_result[ApiAttribute.STATUS] in [
                                ApiStatus.INVALID_TOKEN,
                                ApiStatus.TOKEN_EXPIRED,
                            ]:
                                self._token = None
                                raise SwitchBeeTokenError(
                                    json_result[ApiAttribute.STATUS]
                                )

                            if json_result[ApiAttribute.STATUS] == ApiStatus.OFFLINE:
                                raise SwitchBeeDeviceOfflineError(
                                    f"Central Unit replied with bad status ({json_result[ApiAttribute.STATUS]}): {json_result}"
                                )

                            raise SwitchBeeError(
                                f"Central Unit replied with bad status ({json_result[ApiAttribute.STATUS]}): {json_result}"
                            )
                        else:
                            return json_result
                    except JSONDecodeError:
                        raise SwitchBeeError(f"Unexpected response: {response.read()}")
                else:
                    raise SwitchBeeError(
                        f"Request to the Central Unit failed with status={response.status}"
                    )
        except TimeoutError as exp:
            raise SwitchBeeError(
                "Timed out while waiting for the Central Unit to reply"
            ) from exp

        except ClientConnectorError as exp:
            raise SwitchBeeError("Failed to communicate with the Central Unit") from exp

    async def _login(self) -> None:
        try:
            resp = await self.call(
                ApiCommand.LOGIN,
                {
                    ApiAttribute.USER: self._username,
                    ApiAttribute.PASS: self._password,
                },
            )

        except SwitchBeeError:
            self._token = None
            self._token_expiration = 0
            raise

        super()._update_login(resp)

    async def _send_request(self, command: str, params: Any = {}) -> dict:
        return await self.call(command, params)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass, field
from enum import Enum, unique
from typing import Any, List, final

from ..const import ApiDeviceHardware, ApiDeviceType, ApiStateCommand
from ..utils import timestamp_now


@unique
class DeviceType(Enum):
    """Enum class representing the device's type."""

    Dimmer = ApiDeviceType.DIMMER, "Light"
    Switch = ApiDeviceType.SWITCH, "Switch"
    Shutter = ApiDeviceType.SHUTTER, "Shutter"
    Scenario = ApiDeviceType.SCENARIO, "Scenario"
    Repeater = ApiDeviceType.REPEATER, "Repeater"
    GroupSwitch = ApiDeviceType.GROUP_SWITCH, "Group Switch"
    TwoWay = ApiDeviceType.TWO_WAY, "Two Way"
    TimedPowerSwitch = ApiDeviceType.TIMED_POWER, "Timed Power Switch"
    Thermostat = ApiDeviceType.THERMOSTAT, "Thermostat"
    LockGroup = ApiDeviceType.LOCK_GROUP, "Lock Group"
    TimedSwitch = ApiDeviceType.TIMED_SWITCH, "Timed Switch"
    Somfy = ApiDeviceType.SOMFY, "Somfy"
    IrDevice = ApiDeviceType.IR_DEVICE, "Infra Red Device"
    RollingScenario = ApiDeviceType.ROLLING_SCENARIO, "Rolling Scenario"
    VRFAC = ApiDeviceType.VRF_AC, "VRF Unit"

    def __new__(cls, *args: Any, **kwds: Any):  # type: ignore
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    # ignore the first param since it's already set by __new__
    def __init__(self, _: str, display: str = "") -> None:
        self._display = display

    def __str__(self) -> str:
        return self.display

    # this makes sure that the description is read-only
    @property
    def display(self) -> str:
        return self._display


class HardwareType(Enum):
    Virtual = ApiDeviceHardware.VIRTUAL, "Virtual"
    Dimmable = ApiDeviceHardware.DIMMABLE_SWITCH, "Switch"
    Shutter = ApiDeviceHardware.SHUTTER, "Shutter"
    TimedPowerSwitch = ApiDeviceHardware.TIMED_POWER_SWITCH, "Timed Power Switch"
    Thermostat = ApiDeviceHardware.THERMOSTAT, "CoolSwitch"
    Somfy = ApiDeviceHardware.SOMFY, "Somfy"
    SocketIR = ApiDeviceHardware.SOCKET_IR, "Socket IR"
    StickerSwitch = ApiDeviceHardware.STIKER_SWITCH, "Sticker Switch"
    RegularSwitch = ApiDeviceHardware.REGULAR_SWITCH, "Regular Switch"
    Repeater = ApiDeviceHardware.REPEATER, "Repeater"

    def __new__(cls, *args: Any, **kwds: Any):  # type: ignore
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    # ignore the first param since it's already set by __new__
    def __init__(self, _: str, display: str = "") -> None:
        self._display = display

    def __str__(self) -> str:
        return self.display

    # this makes sure that the description is read-only
    @property
    def display(self) -> str:
        return self._display


@dataclass
class SwitchBeeBaseDevice(ABC):
    id: int
    name: str
    zone: str
    type: DeviceType
    hardware: HardwareType

    def __post_init__(self) -> None:
        """Post initialization, set last_data_update to the instantiation datetime."""
        self.last_data_update = timestamp_now()
        self.unit_id = self.id // 10

    def __hash__(self) -> int:
        return self.id


@dataclass
class SwitchBeeBaseSwitch(ABC):
    _state: str | int | None = field(init=False, default=None)

    @property
    def state(self) -> str | int | None:
        return self._state

    @state.setter
    def state(self, value: str | int) -> None:
        if value in [ApiStateCommand.OFF, 0]:
            self._state = ApiStateCommand.OFF
        elif value in [ApiStateCommand.ON, 100]:
            self._state = ApiStateCommand.ON
        else:  # OFFLINE
            self._state = -1


@dataclass
class SwitchBeeBaseShutter(ABC):
    _position: int | None = field(init=False, repr=False, default=None)

    @property
    def position(self) -> int | None:
        return self._position

    @position.setter
    def position(self, value: str | int) -> None:
        if isinstance(value, int):
            self._position = int(value)
        else:
            if value == ApiStateCommand.ON:
                self._position = 100
            elif value == ApiStateCommand.OFF:
                self._position = 0
            else:  # OFFLINE
                self._position = -1


@dataclass
class SwitchBeeBaseDimmer(ABC):
    _brightness: int = field(init=False)

    @property
    def brightness(self) -> int:
        return self._brightness

    @brightness.setter
    def brightness(self, value: str | int) -> None:
        if isinstance(value, int):
            self._brightness = int(value)
        else:
            if value == ApiStateCommand.ON:
                self._brightness = 100
            elif value == ApiStateCommand.OFF:
                self._brightness = 0
            else:  # OFFLINE
                self._brightness = -1


@dataclass
class SwitchBeeBaseTimer(ABC):
    _minutes_left: int = field(init=False)
    _state: str | int = field(init=False)

    @property
    def state(self) -> str | int:
        return self._state

    @state.setter
    def state(self, value: str | int) -> None:
        if value:
            if value == ApiStateCommand.OFF:
                self._minutes_left = 0
                self._state = value
            else:
                self._minutes_left = int(value)
                self._state = ApiStateCommand.ON

    @property
    def minutes_left(self) -> int:
        return self._minutes_left


@dataclass
class SwitchBeeBaseThermostat(ABC):
    modes: List[str]
    temperature_unit: str
    mode: str = field(init=False)
    fan: str = field(init=False)
    target_temperature: int = field(init=False)
    temperature: int = field(init=False)
    max_temperature: int = 31
    min_temperature: int = 16


@final
@dataclass
class SwitchBeeSwitch(SwitchBeeBaseSwitch, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Switch."""
        if self.type != DeviceType.Switch:
            raise ValueError("only Switch are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeShutter(SwitchBeeBaseShutter, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Shutter."""
        if self.type != DeviceType.Shutter:
            raise ValueError("only Shutter are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeDimmer(SwitchBeeBaseDimmer, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Dimmer."""
        if self.type != DeviceType.Dimmer:
            raise ValueError("only Dimmer are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeTimerSwitch(SwitchBeeBaseTimer, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as TimedPowerSwitch."""
        if self.type != DeviceType.TimedPowerSwitch:
            raise ValueError("only TimedPowerSwitch are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeScenario(SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Scenario."""
        if self.type != DeviceType.Scenario:
            raise ValueError("only Scenario are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeRollingScenario(SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Rolling Scenario."""
        if self.type != DeviceType.RollingScenario:
            raise ValueError("only Rolling Scenario are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeGroupSwitch(SwitchBeeBaseSwitch, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as GroupSwitch."""
        if self.type != DeviceType.GroupSwitch:
            raise ValueError("only GroupSwitch are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeTimedSwitch(SwitchBeeBaseSwitch, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as TimedSwitch."""
        if self.type != DeviceType.TimedSwitch:
            raise ValueError("only TimedSwitch are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeThermostat(
    SwitchBeeBaseThermostat, SwitchBeeBaseSwitch, SwitchBeeBaseDevice
):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Thermostat."""
        if self.type not in [DeviceType.Thermostat, DeviceType.VRFAC]:
            raise ValueError("only Thermostat are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeLockGroup(SwitchBeeBaseSwitch, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Switch."""
        if self.type != DeviceType.LockGroup:
            raise ValueError("only lock group are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeTwoWay(SwitchBeeBaseSwitch, SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as TwoWay."""
        if self.type != DeviceType.TwoWay:
            raise ValueError("only Two Way are allowed")
        super().__post_init__()


@final
@dataclass
class SwitchBeeSomfy(SwitchBeeBaseDevice):
    def __post_init__(self) -> None:
        """Post initialization validate device type category as Scenario."""
        if self.type != DeviceType.Somfy:
            raise ValueError("only Scenario are allowed")
        super().__post_init__()



from __future__ import annotations

from asyncio import Future, TimeoutError, create_task, tasks
from dataclasses import dataclass
from logging import getLogger
from typing import Any, Callable
from .central_unit import CentralUnitAPI, SwitchBeeError

import async_timeout
from aiohttp import ClientSession, ClientWebSocketResponse, WSMessage, WSMsgType
from aiohttp.client_exceptions import ClientError, WSServerHandshakeError
from switchbee.const import ApiAttribute, ApiCommand


logger = getLogger(__name__)


class DeviceConnectionError(Exception):
    pass


class InvalidMessage(Exception):
    pass


class ConnectionClosed(Exception):
    pass


async def receive_json_or_raise(msg: WSMessage) -> dict[str, Any]:
    """Receive json or raise."""
    if msg.type in (WSMsgType.CLOSE, WSMsgType.CLOSED, WSMsgType.CLOSING):
        raise ConnectionClosed("Connection was closed.")

    if msg.type == WSMsgType.ERROR:
        raise InvalidMessage("Received message error")

    if msg.type != WSMsgType.TEXT:
        raise InvalidMessage(f"Received non-Text message: {msg.type}")

    try:
        data: dict[str, Any] = msg.json()
    except ValueError as err:
        raise InvalidMessage(f"Received invalid JSON: {msg.data}") from err

    if status := data.get("status") == "ERROR":
        raise SwitchBeeError("Response Error: %s", data)

    return data


@dataclass
class SessionData:
    """SessionData (src/dst/auth) class."""

    src: str | None
    dst: str | None
    auth: dict[str, Any] | None


class RPCCall:
    """RPCCall class."""

    def __init__(
        self,
        call_id: int,
        command: str | None,
        params: dict[str, Any] | int | list | None,
        token: str | None,
        session: SessionData,
    ):
        """Initialize RPC class."""
        self.auth = session.auth
        self.call_id = call_id
        self.command = command
        self.params = params
        self.token = token
        self.src = session.src
        self.dst = session.dst
        self.resolve: Future = Future()

    @property
    def request_frame(self) -> dict[str, Any]:
        """Request frame."""
        msg: dict[str, Any] = {"commandId": self.call_id, "command": self.command}

        if self.params:
            msg["params"] = self.params

        if self.token:
            msg["token"] = self.token

        return msg


class CentralUnitWsRPC(CentralUnitAPI):
    def __init__(
        self,
        ip_address: str,
        user_name: str,
        password: str,
        aiohttp_session: ClientSession,
        on_notification: Callable | None = None,
    ) -> None:
        super().__init__(ip_address, user_name, password, aiohttp_session)

        self._client: ClientWebSocketResponse | None = None
        self._receive_task: tasks.Task[None] | None = None
        self._calls: dict[int, RPCCall] = {}
        self._on_notification = on_notification
        self._call_id = 0
        self._session = SessionData(f"swb-{id(self)}", None, None)

    @property
    def connected(self) -> bool:
        """Return if we're currently connected."""
        return self._client is not None and not self._client.closed

    @property
    def _next_id(self) -> int:
        self._call_id += 1
        return self._call_id

    def subscribe_updates(self, callback: Callable) -> None:
        self._on_notification = callback

    async def connect(self) -> None:
        if self.connected:
            raise RuntimeError("Already connected")

        logger.debug("Trying to connect to device at %s", self._ip_address)
        try:
            self._client = await self._aiohttp_session.ws_connect(
                f"http://{self._ip_address}:7891"
            )
        except (
            WSServerHandshakeError,
            ClientError,
        ) as err:
            raise DeviceConnectionError(err) from err

        except TimeoutError as err:
            raise DeviceConnectionError(err) from err

        self._receive_task = create_task(self._rx_msgs())
        await self._login()
        await self.fetch_configuration()
        await self.fetch_states()
        logger.info("Connected to %s", self._ip_address)

    async def _send_json(self, data: dict[str, Any]) -> None:
        """Send json frame to device."""
        logger.debug("send(%s): %s", self._ip_address, data)
        assert self._client
        await self._client.send_json(data)

    async def _rx_msgs(self) -> None:
        assert self._client

        while not self._client.closed:
            try:
                msg = await self._client.receive()
                frame = await receive_json_or_raise(msg)
                logger.debug("recv(%s): %s", self._ip_address, frame)
            except InvalidMessage as err:
                logger.error(
                    "Invalid Message from central unit %s: %s", self._ip_address, err
                )
            except ConnectionClosed:
                break

            except SwitchBeeError as err:
                logger.error(
                    "Response error from central unit %s: %s", self._ip_address, err
                )

            if not self._client.closed:
                self.handle_frame(frame)

        logger.debug("Websocket client connection from %s closed", self._ip_address)

        for call_item in self._calls.values():
            call_item.resolve.cancel()
        self._calls.clear()

        if not self._client.closed:
            await self._client.close()

        self._client = None

    def handle_frame(self, frame: dict[str, Any]) -> None:
        """Handle RPC frame."""

        if command_id := frame.get("commandId"):
            # looks like a response
            if command_id not in self._calls:
                logger.warning("Response for an unknown request id: %s", command_id)
                return

            call = self._calls.pop(command_id)
            if not call.resolve.cancelled():
                call.resolve.set_result(frame)

        else:
            if notification_type := frame.get("notificationType"):
                # this is a notification
                logger.debug("Notification %s %s", notification_type, frame)
                # update the devices states based on the event
                self.update_device_state_from_event(frame)
                if self._on_notification:
                    self._on_notification(frame)
            else:
                logger.warning("Invalid frame: %s", frame)

    async def call(
        self,
        command: str | None = None,
        params: dict[str, Any] | int | list | None = None,
        timeout: int = 10,
    ) -> dict[str, Any]:
        """Websocket RPC call."""

        if self._client is None:
            raise RuntimeError("Not connected")

        call = RPCCall(self._next_id, command, params, self._token, self._session)
        self._calls[call.call_id] = call

        try:
            async with async_timeout.timeout(timeout):
                await self._send_json(call.request_frame)
                resp: dict[str, Any] = await call.resolve
        except TimeoutError as exc:
            raise DeviceConnectionError(call) from exc

        logger.debug("%s -> %s", call.params, resp)
        return resp

    async def _send_request(
        self, command: str, params: dict[str, Any] | int | list | None = None
    ) -> dict:
        if not self.connected:
            await self.connect()

        return await self.call(command, params)

    async def _login(self) -> None:

        resp = await self.call(
            ApiCommand.LOGIN,
            {
                ApiAttribute.USER: self._username,
                ApiAttribute.PASS: self._password,
            },
        )

        super()._update_login(resp)

    def update_device_state_from_event(self, push_data: dict) -> None:
        """Update device state from notification data."""

        if ApiAttribute.ID not in push_data:
            logger.error("Received update with no device id: %s", push_data)
            return

        self.update_device_state(
            push_data[ApiAttribute.ID], push_data[ApiAttribute.NEW_VALUE]
        )
